#include "ag.h"
#include "grafo.c"
#include "navega.c"

void inicia(Populacao *pop, Vertice vertices[]){
    int i, j;
    pop->tam = TAM_POP;

    for(i = 0; i < pop->tam; i++){
	   for(j = 0; j < TAM_IND; j++){
	      pop->individuos[i].acoes[j] = rand()%5;
   	   }
	   pop->individuos[i].fitness = avaliaV(vertices, pop->individuos[i].acoes, 0);
    }
}

void mutacao(Individuo *ind){
	int i, a;
	float r = rand()%10;
	r = r/100;
	if(r < 0.05){
	   i = rand()%TAM_IND;
	   a = rand()%5;
	   ind->acoes[i] = a;
	}
}

Individuo selecao(Populacao *pop){
	int i, i1, i2;
	i1 = rand()%TAM_POP;
	i2 = rand()%TAM_POP;
	while(i1 == i2)
	   i2 = rand()%TAM_POP;
	   
	if(pop->individuos[i1].fitness < pop->individuos[i2].fitness)
	   return pop->individuos[i1];
	else
	   return pop->individuos[i2];
}

Individuo cruzamento(Populacao *pop, Vertice vertices[]){
	Individuo pai1, pai2;
	Individuo filho;
	int i, count;
	float r;
	count = 0;

	do{
	   //selecionando os pais (fitness diferentes)
	   pai1 = selecao(pop);
	   pai2 = selecao(pop);
	   while(abs(pai1.fitness - pai2.fitness) == 0 && count < 10){
	      pai2 = selecao(pop);
	      count++;
	   }

	   //gerando o filho
	   for(i = 0; i < TAM_IND; i++){
	      r = rand()%10;
	      r = r/10;
	      if(r > 0.5)
	         filho.acoes[i] = pai1.acoes[i];
	      else
	         filho.acoes[i] = pai2.acoes[i];
	   }

	   //mutação no filho
	   mutacao(&filho);

	   //avaliando o filho
	   filho.fitness = avaliaV(vertices, filho.acoes, 0);

	//se o fitness do filhoo for igual a um dos pais, gera um novo filho
	} while((abs(pai1.fitness - filho.fitness) == 0) || (abs(pai2.fitness - filho.fitness) == 0));
	return filho;
}

void merge(Individuo individuos[], int tam){
  int mid;
  int i, j, k;
  Individuo* tmp;
 
  tmp = (Individuo*) malloc(tam * sizeof(Individuo));
  if (tmp == NULL) {
    exit(1);
  }
 
  mid = tam / 2;
 
  i = 0;
  j = mid;
  k = 0;
  while (i < mid && j < tam) {
    if(individuos[i].fitness < individuos[j].fitness){
      tmp[k] = individuos[i++];
    }
    else {
      tmp[k] = individuos[j++];
    }
    ++k;
  }
 
  if (i == mid) {
    while (j < tam) {
      tmp[k++] = individuos[j++];
    }
  }
  else {
    while (i < mid) {
      tmp[k++] = individuos[i++];
    }
  }
 
  for (i = 0; i < tam; ++i) {
    individuos[i] = tmp[i];
  }
 
  free(tmp);
}
 
void mergeSort(Individuo individuos[], int tam){
  int mid;
 
  if (tam > 1) {
    mid = tam / 2;
    mergeSort(individuos, mid);
    mergeSort(individuos + mid, tam - mid);
    merge(individuos, tam);
  }
}

Individuo melhorFitness(Populacao *pop){
	int i, melhorI = 0;
	float melhor = pop->individuos[melhorI].fitness;

	for(i = 1; i < pop->tam; i++){
	   if(pop->individuos[i].fitness < melhor){
		melhor = pop->individuos[i].fitness;
		melhorI = i;
	   }
	}
	return pop->individuos[melhorI];
}

void imprimirPop(Populacao *pop, Vertice vertices[]){
	int i;
	for(i = 0; i < pop->tam; i++){
	    //printf("i: %d\n", i);
	    avaliaV(vertices, pop->individuos[i].acoes, 0);
	    printf("i: %d; fit: %f\n", i, pop->individuos[i].fitness);
	}
	printf("\n");
}

void copia(Individuo *ind1, Individuo *ind2){
	int i;
	ind2->fitness = ind1->fitness;
	for(i = 0; i < TAM_IND; i++)
	   ind2->acoes[i] = ind1->acoes[i];
	ind2->tam = ind1->tam;
}

Individuo executarAG(Vertice vertices[]){
	int i, g, tam, gmi, count;
	Populacao pop;
	float f, fMedio, somaF;
	Individuo filho, in, melhorI;
	clock_t inicio, fim;	
	
	inicia(&pop, vertices);
	
	melhorI.fitness = 1000;

	count = 0;
	g = 0;
	inicio = clock();
	while(g < NUM_GERACOES && count < 10){
	   mergeSort(pop.individuos, pop.tam);
	   
	   in = melhorFitness(&pop);
	   
	   if(melhorI.fitness == in.fitness)
		count++;
	   else
		count = 0;
	   if(in.fitness < melhorI.fitness){
		copia(&in, &melhorI);
		gmi = g;
	   }
	   
	   avaliaV(vertices, in.acoes, 0);
	   
	   //pop.tam = TAM_POP;
	   pop.tam = 1;
	   for(i = 1; i < TAM_POP; i++){
	      filho = cruzamento(&pop, vertices);
	      //pop.individuos[i + TAM_POP] = filho;
	      pop.individuos[i] = filho;
	      pop.tam++;
	   }
	   g++;
	}
	
	fim = clock();
	
	somaF = 0;
	for(i = 0; i < pop.tam; i++){
	      somaF = somaF + pop.individuos[i].fitness;
	}
	fMedio = (somaF/pop.tam);
	printf("%f;%d;%f\n", in.fitness, (g-1), ((float) (fim - inicio)));

	//avaliaV(vertices, melhorI.acoes, 1);

	return melhorI;
}

int main(){
	int i, g, qtdVertices = 119;
	Individuo in, mIn;
	Vertice vertices[qtdVertices];
	srand(time(NULL));

	inicializaVertices(vertices, qtdVertices);

	mIn.fitness = 10000;
	//printf("best_fitness;generations;time\n");
	for(i = 0; i < 100; i++){
	    //printf("============================================================> Execução: %d\n", i);
	    in = executarAG(vertices);
	    //printf("in: %f\n", in.fitness);
	    if(in.fitness < mIn.fitness){
		copia(&in, &mIn);
	    }
	}
	//printf("mIn: %f\n", mIn.fitness);

	avaliaV(vertices, mIn.acoes, 1);
	for(i = 0; i < TAM_IND; i++)
	   printf("%d ", mIn.acoes[i]);
	//sleep(2);
	//navega(mIn.acoes, TAM_IND);

}

